#!/usr/bin/env python

"""
Moin wiki format parser.

Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.parsing import ParserBase, TokenStream, get_patterns, new_block
from moinformat.serialisers import serialise
from moinformat.tree import Break, DefItem, DefTerm, FontStyle, Heading, \
                            Larger, ListItem, Monospace, Region, Rule, Smaller, \
                            Subscript, Superscript, TableAttr, TableAttrs, \
                            TableCell, TableRow, Text, Underline

# Regular expressions.

syntax = {
    # Page regions:
    "regionstart"   : r"((^\s*)([{]{3,}))",                         # {{{...
    "regionend"     : r"^\s*([}]{3,})",                             # }}}...
    "header"        : r"#!(.*?)\n",                                 # #! char-excl-nl

    # Region contents:
    # Line-oriented patterns:
                      # blank line
    "break"         : r"^(\s*?)\n",
                      # ws... expecting text ::
    "defterm"       : r"^(\s+)(?=.+?::)",
                      # ws... expecting :: ws...
    "defterm_empty" : r"^(\s+)(?=::\s+)",
                      # [ws...] =... ws... expecting headingend
    "heading"       : r"^(\s*)(?P<x>=+)(\s+)(?=.*?\s+(?P=x)\s*\n)",
                      # ws... list-item [ws...]
    "listitem"      : r"^(\s+)(\*)(\s*)",
                      # ws... number-item ws...
    "listitem_num"  : r"^(\s+)(\d+\.)(\s+)",
                      # ws... alpha-item ws...
    "listitem_alpha": r"^(\s+)([aA]\.)(\s+)",
                      # ws... roman-item ws...
    "listitem_roman": r"^(\s+)([iI]\.)(\s+)",
                      # ws... dot-item [ws...]
    "listitem_dot"  : r"^(\s+)(\.)(\s*)",
                      # ||
    "tablerow"      : r"^\|\|",

    # Region contents:
    # Inline patterns:
    "fontstyle"     : r"('{2,6})",
    "larger"        : r"~\+",
    "monospace"     : r"`",
    "rule"          : r"(-----*)",                                  # ----...
    "smaller"       : r"~-",
    "sub"           : r",,",
    "super"         : r"\^",
    "underline"     : r"__",

    # Inline contents:
    "largerend"     : r"\+~",
    "monospaceend"  : r"`",
    "smallerend"    : r"-~",
    "subend"        : r",,",
    "superend"      : r"\^",
    "underlineend"  : r"__",

    # Heading contents:
    "headingend"    : r"(\s+)(=+)(\s*\n)",                          # ws... =... [ws...] nl

    # List contents:
    "deftermend"    : r"::(\s*?\n)",
    "deftermsep"    : r"::(\s+)",
    "listitemend"   : r"^",                                         # next line

    # Table contents:
    "tableattrs"    : r"<",
    "tablecell"     : r"\|\|",
    "tableend"      : r"(\s*?)^",                                   # [ws...] next line

    # Table attributes:
    "tableattrsend" : r">",
    "halign"        : r"([(:)])",
    "valign"        : r"([v^])",
    "colour"        : r"(\#[0-9A-F]{6})",
    "colspan"       : r"-(\d+)",
    "rowspan"       : r"\|(\d+)",
    "width"         : r"(\d+%)",
    "attrname"      : r"((?![-\d])[-\w]+)",                         # not-dash-or-digit dash-or-word-char...
    "attrvalue"     : r"""=(?P<x>['"])(.*?)(?P=x)""",
    }



class Parser(ParserBase):

    "A wiki region parser."

    def __init__(self, formats=None):

        """
        Initialise the parser with any given 'formats' mapping from region type
        names to parser objects.
        """

        formats = {"wiki" : self}
        if formats:
            formats.update(formats)

        ParserBase.__init__(self, formats)

    # Pattern details.

    patterns = get_patterns(syntax)

    table_pattern_names = [
        "attrname", "colour", "colspan", "halign", "rowspan", "tableattrsend",
        "valign", "width"
        ]

    inline_pattern_names = [
        "fontstyle", "larger", "monospace", "smaller", "sub", "super", "underline",
        ]

    region_pattern_names = inline_pattern_names + [
        "break", "heading", "defterm", "defterm_empty", "listitem",
        "listitem_alpha", "listitem_dot", "listitem_num", "listitem_roman",
        "regionstart", "regionend", "rule", "tablerow",
        ]

    def inline_patterns_for(self, name):
        names = self.inline_pattern_names[:]
        names[names.index(name)] = "%send" % name
        return names

    # Principal parser methods.

    def get_items(self, s):

        "Return a sequence of token items for 's'."

        return TokenStream(s, self.patterns)

    def parse(self, s):

        """
        Parse page text 's'. Pages consist of regions delimited by markers.
        """

        items = self.get_items(s)
        region = Region([])

        # Parse page header.

        self.parse_region_header(items, region)

        # Handle pages directly with this parser.
        # Otherwise, test the type and find an appropriate parser.

        if not region.type:
            self.parse_region_content(items, region)
        else:
            self.parse_region_type(items, region)

        return region

    def parse_region_content(self, items, region):

        "Parse the data provided by 'items' to populate a wiki 'region'."

        new_block(region)
        self.parse_region_details(items, region, self.region_pattern_names)

    # Parser methods supporting different page features.

    def parse_attrname(self, items, attrs):

        "Handle an attribute name within 'attrs'."

        name = items.read_match()
        attr = TableAttr(name)

        preceding = items.read_until(["attrvalue"], False)
        if preceding == "":
            attr.quote = items.read_match(1)
            attr.value = items.read_match(2)

        attrs.append(attr)

    def parse_break(self, items, region):

        "Handle a paragraph break within 'region'."

        region.add(Break())
        new_block(region)

    def parse_defitem(self, items, region, extra=""):

        "Handle a definition item within 'region'."

        pad = items.read_match(1)
        item = DefItem([], pad, extra)
        self.parse_region_details(items, item, ["listitemend"])
        region.add(item)
        new_block(region)

    def parse_defterm(self, items, region):

        "Handle a definition term within 'region'."

        pad = items.read_match(1)
        term = DefTerm([], pad)
        self.parse_region_details(items, term, ["deftermend", "deftermsep"])
        region.add(term)
        if items.matching == "deftermsep":
            self.parse_defitem(items, region)

    def parse_defterm_empty(self, items, region):

        "Handle an empty definition term within 'region'."

        extra = items.read_match(1)
        self.parse_region_details(items, region, ["deftermsep"])
        self.parse_defitem(items, region, extra)

    def parse_fontstyle(self, items, region):

        "Handle emphasis and strong styles."

        n = len(items.read_match(1))

        # Handle endings.

        if isinstance(region, FontStyle):
            emphasis = n in (2, 4, 5)
            strong = n in (3, 5, 6)
            active = True

            if region.emphasis and emphasis:
                active = region.close_emphasis()
                n -= 2
            if region.strong and strong:
                active = region.close_strong()
                n -= 3

            if not active:
                if n:
                    items.rewind(n)
                raise StopIteration

            elif not n:
                return

        # Handle new styles.

        emphasis = n in (2, 4, 5)
        strong = n in (3, 5, 6)
        double = n in (4, 6)

        span = FontStyle([], emphasis, strong)
        if not double:
            self.parse_region_details(items, span, self.inline_pattern_names)
        region.append_inline(span)

    def parse_halign(self, items, attrs):

        "Handle horizontal alignment within 'attrs'."

        value = items.read_match()
        attr = TableAttr("halign", value == "(" and "left" or value == ")" and "right" or "center", True)
        attrs.append(attr)

    def parse_heading(self, items, region):

        "Handle a heading."

        start_extra = items.read_match(1)
        level = len(items.read_match(2))
        start_pad = items.read_match(3)
        heading = Heading([], level, start_extra, start_pad)
        self.parse_region_details(items, heading, ["headingend"] + self.inline_pattern_names)
        region.add(heading)
        new_block(region)

    def parse_heading_end(self, items, heading):

        "Handle the end of a heading."

        level = len(items.read_match(2))
        if heading.level == level:
            heading.end_pad = items.read_match(1)
            heading.end_extra = items.read_match(3)
            raise StopIteration

    def parse_listitem(self, items, region):

        "Handle a list item marker within 'region'."

        indent = len(items.read_match(1))
        marker = items.read_match(2)
        space = items.read_match(3)
        item = ListItem([], indent, marker, space)
        self.parse_region_details(items, item, ["listitemend"])
        region.add(item)
        new_block(region)

    def parse_rule(self, items, region):

        "Handle a horizontal rule within 'region'."

        length = len(items.read_match(1))
        rule = Rule(length)
        region.add(rule)
        new_block(region)

    def parse_section(self, items, region):

        "Handle the start of a new section within 'region'."

        # Parse the section and start a new block after the section.

        indent = len(items.read_match(2))
        level = len(items.read_match(3))
        region.add(self.parse_region(items, level, indent))
        new_block(region)

    def parse_section_end(self, items, region):

        "Handle the end of a new section within 'region'."

        feature = items.read_match()
        if region.have_end(feature):
            raise StopIteration
        else:
            region.append_inline(Text(feature))

    def parse_table_attrs(self, items, cell):

        "Handle the start of table attributes within 'cell'."

        attrs = TableAttrs([])
        self.parse_region_details(items, attrs, self.table_pattern_names)

        # Test the validity of the attributes.

        last = None

        for node in attrs.nodes:

            # Text separator nodes must be whitespace.

            if isinstance(node, Text):
                if node.s.strip():
                    break

            # Named attributes must be preceded by space if not the first.

            elif last and not node.concise and not isinstance(last, Text):
                break

            last = node

        # All nodes were valid: preserve the collection.

        else:
            cell.attrs = attrs
            return

        # Invalid nodes were found: serialise the attributes as text.

        cell.append_inline(Text(serialise(attrs)))

    def parse_table_row(self, items, region):

        "Handle the start of a table row within 'region'."

        row = TableRow([])

        while True:
            cell = TableCell([])
            self.parse_region_details(items, cell, ["tableattrs", "tablecell", "tableend"])

            # Handle the end of the row.

            if items.matching == "tableend":
                trailing = items.read_match()

                # If the cell was started but not finished, convert the row into text.

                if not row.nodes or not cell.empty():
                    for node in row.nodes:
                        region.append_inline(Text(serialise(node)))
                    region.append_inline(Text(serialise(cell)))
                    region.append_inline(Text(trailing))

                    new_block(region)
                    return

                # Append the final cell, if not empty.

                else:
                    row.trailing = trailing

                    if not cell.empty():
                        row.append(cell)
                    break

            # A cell separator has been found.

            row.append(cell)

        region.add(row)
        new_block(region)

    def parse_valign(self, items, attrs):

        "Handle vertical alignment within 'attrs'."

        value = items.read_match()
        attr = TableAttr("valign", value == "^" and "top" or "bottom", True)
        attrs.append(attr)



    # Inline formatting handlers.

    def parse_inline(self, items, region, cls, pattern_name):

        "Handle an inline region."

        span = cls([])
        self.parse_region_details(items, span, self.inline_patterns_for(pattern_name))
        region.append_inline(span)

    def parse_larger(self, items, region):
        self.parse_inline(items, region, Larger, "larger")

    def parse_monospace(self, items, region):
        self.parse_inline(items, region, Monospace, "monospace")

    def parse_smaller(self, items, region):
        self.parse_inline(items, region, Smaller, "smaller")

    def parse_sub(self, items, region):
        self.parse_inline(items, region, Subscript, "sub")

    def parse_super(self, items, region):
        self.parse_inline(items, region, Superscript, "super")

    def parse_underline(self, items, region):
        self.parse_inline(items, region, Underline, "underline")



    # Table attribute handlers.

    def parse_table_attr(self, items, attrs, pattern_name):

        "Handle a table attribute."

        value = items.read_match()
        attrs.append(TableAttr(pattern_name, value, True))

    def parse_colour(self, items, cell):
        self.parse_table_attr(items, cell, "colour")

    def parse_colspan(self, items, cell):
        self.parse_table_attr(items, cell, "colspan")

    def parse_rowspan(self, items, cell):
        self.parse_table_attr(items, cell, "rowspan")

    def parse_width(self, items, cell):
        self.parse_table_attr(items, cell, "width")



    # Pattern handlers.

    end_region = ParserBase.end_region

    handlers = {
        None : end_region,
        "attrname" : parse_attrname,
        "break" : parse_break,
        "colour" : parse_colour,
        "colspan" : parse_colspan,
        "defterm" : parse_defterm,
        "defterm_empty" : parse_defterm_empty,
        "deftermend" : end_region,
        "deftermsep" : end_region,
        "fontstyle" : parse_fontstyle,
        "halign" : parse_halign,
        "heading" : parse_heading,
        "headingend" : parse_heading_end,
        "larger" : parse_larger,
        "largerend" : end_region,
        "listitemend" : end_region,
        "listitem" : parse_listitem,
        "listitem_alpha" : parse_listitem,
        "listitem_dot" : parse_listitem,
        "listitem_num" : parse_listitem,
        "listitem_roman" : parse_listitem,
        "monospace" : parse_monospace,
        "monospaceend" : end_region,
        "regionstart" : parse_section,
        "regionend" : parse_section_end,
        "rowspan" : parse_rowspan,
        "rule" : parse_rule,
        "smaller" : parse_smaller,
        "smallerend" : end_region,
        "sub" : parse_sub,
        "subend" : end_region,
        "super" : parse_super,
        "superend" : end_region,
        "tableattrs" : parse_table_attrs,
        "tableattrsend" : end_region,
        "tablerow" : parse_table_row,
        "tablecell" : end_region,
        "tableend" : end_region,
        "underline" : parse_underline,
        "underlineend" : end_region,
        "valign" : parse_valign,
        "width" : parse_width,
        }



# Top-level functions.

def parse(s, formats=None):
    return Parser(formats).parse(s)

# vim: tabstop=4 expandtab shiftwidth=4
